# -*- coding: utf-8 -*-
# -*- mode: python -*-
"""spike detection routines

Copyright (C) 2013 Dan Meliza <dmeliza@gmail.com>
Created Wed Jul 24 09:26:36 2013
"""
from .spikes import detector, peaks, subthreshold
from .tools import realign_spikes, filter_times

__version__ = "2.0.2"
__all__ = ["detector", "peaks", "subthreshold", "realign_spikes", "filter_times"]
